//
// Oliver Wilson <oliver.wilson@ed.ac.uk>
//

#ifndef CLIENT_H
#define CLIENT_H

#include <google/sparse_hash_map>

#include "ProtocolGroup.h"
#include "TableDirectory.h"
#include "TableProtocol.h"
#include "ticpp.h"

namespace LDHT {

class Quantiser;
class FactoryCollection;
class ServerLocator;
class NewNgram;

struct BackoffCacheEntry {
    enum FlagMasks {
        PendingMask =        0x01,
        ExistsMask =         0x02,
        CalculatedFullMask = 0x04
    };

    uint8_t flags;
    int context_tag;
    int backoff_tag;
    float this_logprob;
    float this_backoff;
    float full_logprob;

    BackoffCacheEntry() {
        this->flags = 0;
    }

    inline void setPendingFlag() {
        this->flags |= PendingMask;
    }
    inline void clearPendingFlag() {
        this->flags &= (!PendingMask);
    }
    inline bool getPendingFlag() {
        return this->flags & PendingMask;
    }

    inline void setExistsFlag() {
        this->flags |= ExistsMask;
    }
    inline void clearExistsFlag() {
        this->flags &= (!ExistsMask);
    }
    inline bool getExistsFlag() {
        return this->flags & ExistsMask;
    }

    inline void setCalculatedFullFlag() {
        this->flags |= CalculatedFullMask;
    }
    inline void clearCalculatedFullFlag() {
        this->flags &= (!CalculatedFullMask);
    }
    inline bool getCalculatedFullFlag() {
        return this->flags & CalculatedFullMask;
    }

    inline void clearAllFlags() {
        this->flags = 0;
    }
};

struct IdentityHasher {
    inline size_t operator()(const uint64_t key) const {
        return key;
    }
};

class Client : public TableProtocolCallback {
public:
    Client();
    ~Client();

    // Configure a Client and all its dependencies from the xml files
    // pointed to by ldht_config_path and ldhtlm_config_path.
    virtual void fromXmlFiles(FactoryCollection& factory_collection,
                              std::string ldht_config_path,
                              std::string ldhtlm_config_path);
    virtual void fromXml(FactoryCollection& factory_collection,
                         ticpp::Element& config_element,
                         ticpp::Element& lm_element);

    // requestNgramString() and requestNgram() are the main entry points
    // to requesting an ngram.
    //
    // The int returned is the tag which should later be given to
    // getNgramScore() after awaitResponses() has been called. The tag
    // is valid, and the ngram's score available until clearCache() is
    // called. After this point, the ngram must be re-requested.
    int requestNgramString(const char* ngram_str);
    int requestNgram(NewNgram& ngram);

    // If the ngram being requested is a one word extension of a
    // previousely requested ngram, then a good deal of work can be
    // avoided by passing in the tag of the previous ngram's backoff,
    // as this is the context for it's rightward extension.
    int requestNgramWithContextTag(const char* gram,
                                   NewNgram& ngram,
                                   int context_tag,
                                   int* next_context_tag);

    // Given a tag returned from any of the request methods above,
    // getNgramScore() returns the ngram's score including backoff
    // penalty.
    float getNgramScore(int ngram_tag);

    // Insert the provided ngram string and values into the model.
    // ngram_str should be a null terminated string of grams separated
    // by whitespace. Separating this into constituent grams is fairly
    // robust, but no guarentees.
    void insertNgram(const char* ngram_str, float logprob, float backoff);

    // Methods to access the various statistics kept about the operation
    // of the Client.
    // TODO(wilson): name sanitation.
    uint64_t getLatencyTicks();
    uint64_t getInsertionAttempts();
    uint64_t getInsertionSuccesses();
    uint64_t getInsertionUpdates();
    uint64_t getNumNgramsSubmitted();
    uint64_t getNumNgramsRequested();
    void resetNumNgramsRequested();
    uint64_t getNumWritesToTransport();
    uint64_t getInferenceCount();
    void resetInferenceCount();
    uint64_t getCacheHitCount();
    void resetCacheHitCount();
    uint64_t getKeyNotFoundCount();
    void resetKeyNotFoundCount();

    // You won't really need anything below here.

    void addTableProtocol(TableProtocol* table_protocol);
    void setServerLocator(ServerLocator* server_locator);

    // Internal methods for retrieving and calculating ngram log
    // probabilities.
    int requestNgramWithContextTag(uint64_t ngram_fp,
                                   NewNgram& ngram,
                                   int context_tag,
                                   int* next_context_tag,
                                   int start_idx,
                                   int end_idx);
    int requestNgram(uint64_t ngram_fp,
                     NewNgram& ngram,
                     int start_idx,
                     int end_idx);
    int getCachedOrRequest(uint64_t ngram_fp,
                           NewNgram& ngram,
                           int start_idx,
                           int end_idx);
    void insertNgram(NewNgram& ngram, float logprob, float backoff);

    // Value interpretation methods.
    void setNumLogprobBits(int num_logprob_bits);
    int getNumLogprobBits();
    void setNumBackoffBits(int num_backoff_bits);
    int getNumBackoffBits();
    void setLogprobQuantiser(Quantiser* quantiser);
    void setBackoffQuantiser(Quantiser* quantiser);
    void setUnkLogprob(float unk_logprob);
    void setUnkBackoff(float unk_backoff);
    void unpackValue(uint64_t value, float* logprob_ret, float* backoff_ret);
    void packValue(float logprob, float backoff, uint64_t* value_ret);

    // Cache access methods.
    BackoffCacheEntry* getCacheEntry(int tag);
    void clearCache();
    int getKeyTag(uint64_t key);
    void setKeyTag(uint64_t key, int tag);

    // Low level methods for issuing requests and handling responses.
    int getLmTableId();
    void setLmTableId(int table_id);
    void get(uint64_t tag, uint64_t key);
    void set(uint64_t tag, uint64_t key, uint64_t value);
    void resetTag();
    virtual void awaitResponses();
    virtual void write();
    virtual void onOpGetResponse(TableProtocol::OpFlags flags,
                                 uint64_t tag,
                                 uint64_t value);
    virtual void onOpSetResponse(TableProtocol::OpFlags flags,
                                 uint64_t tag);
    virtual void onOpSyncResponse(TableProtocol::OpFlags flags,
                                  uint64_t tag);
    virtual void onOpWriteResponse(TableProtocol::OpFlags flags,
                                   uint64_t tag);
    int getNumServers();

    // This was just for debugging integration with Moses but might be
    // useful elsewhere.
    #if 0
    void cacheEntriesToStream(BackoffCacheEntry& entry, std::ostream& stream, std::string indent) {
        stream
        << indent << "pending:      " << entry.getPendingFlag() << std::endl
        << indent << "exists:       " << entry.getExistsFlag() << std::endl
        << indent << "calc full:    " << entry.getCalculatedFullFlag() << std::endl
        << indent << "context_tag:  " << entry.context_tag << std::endl
        << indent << "backoff_tag:  " << entry.backoff_tag << std::endl
        << indent << "this_logprob: " << entry.this_logprob << std::endl
        << indent << "this_backoff: " << entry.this_backoff << std::endl
        << indent << "full_logprob: " << entry.full_logprob << std::endl
        << indent << "context entry: " << std::endl;
        if (entry.context_tag != 0) {
            BackoffCacheEntry* context_entry = getCacheEntry(entry.context_tag);
            cacheEntriesToStream(*context_entry, stream, indent + "    ");
        }
        stream << indent << "backoff entry: " << std::endl;
        if (entry.backoff_tag != 0) {
            BackoffCacheEntry* backoff_entry = getCacheEntry(entry.backoff_tag);
            cacheEntriesToStream(*backoff_entry, stream, indent + "    ");
        }
    }
    #endif

protected:
    TableDirectory m_table_directory;

private:
    // TODO(wilson): The first 3 members belong in a separate class
    // that implements the TableProtocol interface, and is set via
    // setTableProtocol(), rather than addTableProtocol().
    std::vector<TableProtocol*> m_table_protocols;
    ServerLocator* m_server_locator;
    ProtocolGroup m_protocol_group;
    int m_lm_table_id;
    int m_get_tag;
    int m_set_tag;
    int m_sync_tag;
    int m_outstanding_responses;

    // TODO(wilson): Cache could also maybe do with being pulled into
    // a separate class. It should also variable size.
    BackoffCacheEntry* m_cache;
    typedef google::dense_hash_map<uint64_t, int, IdentityHasher> KeyTagMap;
    KeyTagMap m_key_tag_map;

    // Members for dealing with interpreting the raw values in the table
    // into log probabilities and backoff weights.
    Quantiser* m_logprob_quantiser;
    Quantiser* m_backoff_quantiser;
    int m_num_logprob_bits;
    int m_num_backoff_bits;
    uint64_t m_logprob_unpacking_mask;
    uint64_t m_backoff_unpacking_mask;

    // Counters for various operating statistics. Their uint64_t-ness is
    // required because they tend to get big quite quickly.
    // TODO(wilson): name sanitation.
    uint64_t m_gets;
    uint64_t m_insertion_attempts;
    uint64_t m_insertion_successes;
    uint64_t m_insertion_updates;
    uint64_t m_num_ngrams_submitted;
    uint64_t m_num_ngrams_requested;
    uint64_t m_inference_count;
    uint64_t m_cache_hit_count;
    uint64_t m_key_not_found_count;

};

}  // namespace LDHT.

#endif  // CLIENT_H

