//
// Oliver Wilson <oliver.wilson@ed.ac.uk>
//

#ifndef PROTOCOLGROUP_H
#define PROTOCOLGROUP_H

#include <pthread.h>
#include <stdint.h>

#include <google/dense_hash_map>

namespace LDHT {

class Protocol;

// ProtocolGroup is used to monitor a set of Protocol objects to detect
// when they are ready to perform some sort of IO.
//
// It does this through monitoring the file descriptors of the Protocols
// returned by their getWaitFD() methods and performing a select() on
// them.
//
// On calling getReadyProtocol(), all Protocols in the group are checked
// for their shouldMonitor() flag. If this is set, their file
// descriptors will be entered into the select() call. If not, they
// won't be added.
//
// Changing the monitor status of a Protocol will not result in it
// immediately being monitored by the group. Instead, refresh() must
// be called to setup a new call to select().
//
// Interrupting select() is performed by it monitoring an internal file
// descriptor created by pipe(). Another thread can call refresh() on
// the group, which will write to this internal file descriptor causing
// the select to terminate.
//
// In the absence of any other Protocol being ready, getReadyProtocol()
// will determine that refresh() was called and so will refresh its
// set of fds without returning to its caller. Note that if a timeout
// other than 0 was requested, it will start again for scratch. So in
// some cases the caller of getReadyProtocol() may wait longer than the
// timeout they specified, possibly indefinitely.
//
// TODO(wilson): the timeout should be updated. Perhaps make a timeout
// specific method to avoid any overhead.
class ProtocolGroup {
public:
    ProtocolGroup();
    virtual ~ProtocolGroup();

    Protocol* getReadyProtocol(int timeout_usec);
    void addProtocol(Protocol*);
    void removeProtocol(Protocol*);
    void refresh();
    void monitor();
    void monitorOnce();
    uint64_t getLatencyTicks();

private:
    google::dense_hash_map<int, Protocol*> m_fd_protocol_map;
    pthread_mutex_t m_refresh_mutex;
    pthread_mutex_t m_refresh_finish_mutex;
    pthread_cond_t m_refresh_finish_cond;
    int m_refresh_wr_fd;
    int m_refresh_rd_fd;
    uint64_t m_latency_ticks;

};

}  // namespace LDHT.

#endif  // PROTOCOLGROUP_H

