//
// Oliver Wilson <oliver.wilson@ed.ac.uk>
//

#ifndef SERVER_H
#define SERVER_H

#include <vector>

#include "Logic.h"
#include "ProtocolGroup.h"
#include "ProtocolQueue.h"
#include "TableProtocol.h"

namespace LDHT {

class Server;

class ProtocolQueue;

class ServerOpRunner : public Logic {
public:
    explicit ServerOpRunner(int id,
                            ProtocolQueue* protocol_queue,
                            ProtocolGroup* protocol_group);
    virtual ~ServerOpRunner();

    virtual void run();

private:
    int m_id;
    ProtocolQueue* m_protocol_queue;
    ProtocolGroup* m_protocol_group;

};

class AcceptorProtocol : public Protocol {
public:
    AcceptorProtocol(int listen_socket,
                     Server* server,
                     ProtocolGroup* protocol_group);

    virtual void onReady();
    virtual int getWaitFD();
    virtual bool ready();
    virtual bool disconnected();

private:
    int m_listen_socket;
    Server* m_server;
    ProtocolGroup* m_protocol_group;

};

class ServerStateChangeCallback : public StateChangeCallback {
public:
    ServerStateChangeCallback(TableProtocol* table_protocol,
                              ProtocolQueue* protocol_queue,
                              ProtocolGroup* protocol_group);

    virtual void onReady();
    virtual void onDisconnect();

private:
    TableProtocol* m_table_protocol;
    ProtocolQueue* m_protocol_queue;
    ProtocolGroup* m_protocol_group;
};

class TableRange;
class TableChunk;
class Thread;
class FactoryCollection;

//
// Memory Allocation:
//
//   o Any memory allocated by buildServer() becomes the responsibility
//     of the Server instance to deallocate.
//
//   o Any memory allocated by build methods called by buildServer()
//     becomes the responsibility of the objects being built by those
//     methods.
//
// For example, Server will be responsible for deleting the TableRanges
// allocated by buildTableRange() in buildServer(), but TableRange will
// be responsible for deleting TableChunks allocated in buildTableRange.
// 
class Server : public TableProtocolCallback {
public:
    static const int k_default_port;
    static const int k_default_num_threads;

    Server();
    virtual ~Server();

    void fromXml(FactoryCollection& factory_collection,
                 ticpp::Element& server_config,
                 int num_servers,
                 std::vector<ticpp::Element>& table_configs);
    void rangesFromXml(FactoryCollection& factory_collection,
                       std::vector<ticpp::Element>& table_elements,
                       int num_servers,
                       int server_index);
    void addTableRange(TableRange* table_range);
    int getListenSocket();
    void setListenSocket(int socket);
    void listenLoop();
    TableRange* getTableRange(TableIndex table_index);
    TableChunk* getTableChunkForKey(TableIndex table_index, uint64_t key);
    ProtocolQueue* getProtocolQueue();

    virtual TableProtocol::OpFlags onOpGet(TableIndex table_index,
                                           uint64_t tag,
                                           uint64_t key,
                                           uint64_t* value_ret);
    virtual TableProtocol::OpFlags onOpSet(TableIndex table_index,
                                           uint64_t tag,
                                           uint64_t key,
                                           uint64_t value);
    virtual TableProtocol::OpFlags onOpIncrement(TableIndex table_index,
                                                 uint64_t tag,
                                                 uint64_t key,
                                                 uint64_t increment);
    virtual TableProtocol::OpFlags onOpIncrementApprox(TableIndex table_index,
                                                       uint64_t tag,
                                                       uint64_t key,
                                                       uint64_t increment);
    virtual TableProtocol::OpFlags onOpSync(TableIndex table_index,
                                            uint64_t tag);
    virtual TableProtocol::OpFlags onOpWrite(TableIndex table_index,
                                             uint64_t tag);

private:
    int m_listen_socket;
    std::vector<TableRange*> m_table_ranges;
    ProtocolGroup m_protocol_group;
    ProtocolQueue m_protocol_queue;
    std::vector<Thread*> m_op_runner_threads;
    AcceptorProtocol* m_acceptor_protocol;

};

}  // namespace LDHT.

#endif  // SERVER_H

