//
// Oliver Wilson <oliver.wilson@ed.ac.uk>
//

#ifndef TABLEPROTOCOL_H
#define TABLEPROTOCOL_H

#include "Protocol.h"
#include "TypeDef.h"
#include "VarIntStream.h"

namespace LDHT {

class TableProtocolCallback;

class TableProtocol : public Protocol {
public:
    enum OpCode {
        OpCodeNop = 1,
        OpCodeNopResponse,
        OpCodeGet,
        OpCodeGetResponse,
        OpCodeSet,
        OpCodeSetResponse,
        OpCodeIncrement,
        OpCodeIncrementResponse,
        OpCodeIncrementApprox,
        OpCodeIncrementApproxResponse,
        OpCodeSync,
        OpCodeSyncResponse,
        OpCodeWrite,
        OpCodeWriteResponse,
        OpCodeDisconnect,
        OpCodeMax
    };
    enum OpFlags {
        OpFlagNotSupported =    1 << 0,
        OpFlagKeyFound =        1 << 1,
        OpFlagKeyInserted =     1 << 2,
        OpFlagKeyUpdated =      1 << 3,
        OpFlagKeyIncremented =  1 << 4,
        OpFlagValueOverflow =   1 << 5,
        OpFlagSyncSuccessful =  1 << 6,
        OpFlagWriteSuccessful = 1 << 7
    };
    class BufferWriteCallbackNumOps : public BufferWriteCallback {
    public:
        BufferWriteCallbackNumOps();
        BufferWriteCallbackNumOps(TableProtocol*);
        virtual ~BufferWriteCallbackNumOps();
        virtual void preBufferWrite();
        virtual void postBufferWrite();

    private:
        TableProtocol* m_protocol;
    };

    TableProtocol();
    explicit TableProtocol(VarIntStream*);
    virtual ~TableProtocol();

    BufferWriteCallbackNumOps* getBufferWriteCallback();
    void setBufferWriteCallback(BufferWriteCallbackNumOps* buffer_callback);
    void flush();
    bool ready();
    bool disconnected();
    void initialiseOp(OpCode op_code, TableIndex table_index);
    void finaliseCurrentOp();
    void reinitialiseLastOp();
    void processNextBatchOfOps();
    VarIntStream* getVarIntStream();
    virtual int getWaitFD();

    virtual void nop(TableIndex table_index, uint64_t tag);
    virtual void get(TableIndex table_index, uint64_t tag, uint64_t key);
    virtual void set(TableIndex table_index,
                     uint64_t tag,
                     uint64_t key,
                     uint64_t value);
    virtual void increment(TableIndex table_index,
                           uint64_t tag,
                           uint64_t key,
                           uint64_t increment);
    virtual void incrementApprox(TableIndex table_index,
                                 uint64_t tag,
                                 uint64_t key,
                                 uint64_t increment);
    virtual void sync(TableIndex table_index, uint64_t tag);
    virtual void write(TableIndex, uint64_t tag);
    void handleOpNop(TableIndex table_index);
    void handleOpNopResponse();
    void handleOpGet(TableIndex table_index);
    void handleOpGetResponse();
    void handleOpSet(TableIndex table_index);
    void handleOpSetResponse();
    void handleOpIncrement(TableIndex table_index);
    void handleOpIncrementResponse();
    void handleOpIncrementApprox(TableIndex table_index);
    void handleOpIncrementApproxResponse();
    void handleOpSync(TableIndex table_index);
    void handleOpSyncResponse();
    void handleOpWrite(TableIndex table_index);
    void handleOpWriteResponse();

    void setCallback(TableProtocolCallback*);

private:
    VarIntStream* m_var_int_stream;
    TableProtocolCallback* m_callback;
    OpCode m_last_op_code;
    TableIndex m_last_table_index;
    size_t m_num_ops;
    uint8_t* m_num_ops_buffer_loc;
    BufferWriteCallbackNumOps* m_buffer_write_callback;
    bool m_sync_pending;
};

class TableProtocolCallback {
public:
    virtual ~TableProtocolCallback();

    virtual TableProtocol::OpFlags onOpNop(TableIndex table_index,
                                           uint64_t tag);
    virtual TableProtocol::OpFlags onOpGet(TableIndex table_index,
                                           uint64_t tag,
                                           uint64_t key,
                                           uint64_t* value_ret);
    virtual TableProtocol::OpFlags onOpSet(TableIndex table_index,
                                           uint64_t tag,
                                           uint64_t key,
                                           uint64_t value);
    virtual TableProtocol::OpFlags onOpIncrement(TableIndex table_index,
                                                 uint64_t tag,
                                                 uint64_t key,
                                                 uint64_t increment);
    virtual TableProtocol::OpFlags onOpIncrementApprox(TableIndex table_index,
                                                       uint64_t tag,
                                                       uint64_t key,
                                                       uint64_t increment);
    virtual TableProtocol::OpFlags onOpSync(TableIndex table_index,
                                            uint64_t tag);
    virtual TableProtocol::OpFlags onOpWrite(TableIndex table_index,
                                             uint64_t tag);

    virtual void onOpNopResponse(TableProtocol::OpFlags flags, uint64_t tag);
    virtual void onOpGetResponse(TableProtocol::OpFlags flags,
                                 uint64_t tag,
                                 uint64_t value);
    virtual void onOpSetResponse(TableProtocol::OpFlags flags,
                                 uint64_t tag);
    virtual void onOpIncrementResponse(TableProtocol::OpFlags flags,
                                       uint64_t tag);
    virtual void onOpIncrementApproxResponse(TableProtocol::OpFlags flags,
                                             uint64_t tag);
    virtual void onOpSyncResponse(TableProtocol::OpFlags flags, uint64_t tag);
    virtual void onOpWriteResponse(TableProtocol::OpFlags flags, uint64_t tag);
};

}  // namespace LDHT.

#endif  // TABLEPROTOCOL_H

